from typing import Literal, Optional

from pydantic import BaseModel


class Context(BaseModel):
    """
    Represent the context of the current web call.

    **Attributes:**

    * **extension_id** - id of this extension.
    * **environment_id** - id of the environment into which this extension is running.
    * **environment_type** - type of the environment into which this extension is running.
    * **installation_id** - id of the installation object owned by the current caller.
    * **user_id** - id of the user or service user that is doing the call.
    * **account_id** - id of the account that is doing the call.
    * **account_role** - role of the account that is doing the call.
    * **call_source** - it can be `ui` if the call came from the user interface or `api`
        if it is an API call.
    * **call_type** - it can be `admin` if the call came from the same account
        that own the extension otherwise `user`.
    """
    extension_id: Optional[str]
    environment_id: Optional[str]
    environment_type: Optional[Literal['development', 'test', 'production']]
    installation_id: Optional[str] = None
    user_id: Optional[str]
    account_id: Optional[str]
    account_role: Optional[str]
    call_source: Optional[Literal['ui', 'api']]
    call_type: Optional[Literal['admin', 'user']]



import os

from fastapi import Request
from fastapi.responses import JSONResponse, PlainTextResponse

from connect.client import ClientError


def obfuscate_header(key, value):
    if key in ('authorization', 'authentication'):
        if value.startswith('ApiKey '):
            return value.split(':')[0] + ':' + '*' * 10
        else:
            return '*' * 20
    if key in ('cookie', 'set-cookie') and 'api_key="' in value:
        start_idx = value.index('api_key="') + len('api_key="')
        end_idx = value.index('"', start_idx)
        return f'{value[0:start_idx + 2]}******{value[end_idx - 2:]}'
    return value


def client_error_exception_handler(request: Request, exc: ClientError):
    status_code = exc.status_code or 500
    if not exc.error_code:
        return PlainTextResponse(status_code=status_code, content=str(exc))
    else:
        return JSONResponse(
            status_code=status_code,
            content={
                'error_code': exc.error_code,
                'errors': exc.errors,
            },
        )


def get_correlation_id(connect_correlation_id):
    if type(connect_correlation_id) != str:
        return None
    operation_id = connect_correlation_id[3:34]
    span_id = os.urandom(8).hex()
    return f'00-{operation_id}-{span_id}-01'



from connect.eaas.core.validation.validators.anvilapp import validate_anvilapp  # noqa
from connect.eaas.core.validation.validators.base import (  # noqa
    validate_docker_compose_yml,
    validate_extension_json,
    validate_pyproject_toml,
    validate_variables,
)
from connect.eaas.core.validation.validators.eventsapp import validate_eventsapp  # noqa
from connect.eaas.core.validation.validators.webapp import validate_webapp  # noqa


def get_validators():
    return [
        validate_pyproject_toml,
        validate_extension_json,
        validate_docker_compose_yml,
        validate_eventsapp,
        validate_anvilapp,
        validate_webapp,
        validate_variables,
    ]



from connect.eaas.core.enums import ResultType


class _Response:
    def __init__(self, status, output=None):
        self.status = status
        self.output = output

    @classmethod
    def done(cls, *args, **kwargs):
        """
        Create a new response as the task has been
        successfully processed.
        """
        return cls(ResultType.SUCCESS)

    @classmethod
    def fail(cls, output=None):
        """
        Returns a response as the task has been
        successfully processed.

        **Parameters:**

        * **output** - Optional output message to set the reason of failure
            within the task object.
        """
        return cls(ResultType.FAIL, output=output)


class BackgroundResponse(_Response):
    """
    Returns the result of a background event processing.
    """

    def __init__(self, status, countdown=30, output=None):
        super().__init__(status, output)
        self.countdown = 30 if countdown < 30 else countdown

    @classmethod
    def skip(cls, output=None):
        """
        Returns a response as the extension wants to skip the processing
        of the received task.

        **Parameters:**

        * **output** - Optional output message to set the reason of why this
            task has been skipped.
        """
        return cls(ResultType.SKIP, output=output)

    @classmethod
    def reschedule(cls, countdown=30):
        """
        Returns a response as the extension wants to reschedule this task.

        **Parameters:**

        * **countdown** - Optional amount of seconds before next delivery
        of this task (default to 30 seconds).
        """
        return cls(ResultType.RESCHEDULE, countdown=countdown)

    @classmethod
    def slow_process_reschedule(cls, countdown=300):
        """
        Returns a response as the extension wants to reschedule this task.

        **Parameters:**

        * **countdown** - Optional amount of seconds before next delivery
        of this task (default to 300 seconds).

        !!! note
            The minumum amount of seconds to wait before this task will
            be redelivered is 300 seconds (5 minutes).
        """
        return cls(
            ResultType.RESCHEDULE,
            countdown=300 if countdown < 300 else countdown,
        )


class ProcessingResponse(BackgroundResponse):
    pass


class InteractiveResponse(_Response):
    def __init__(self, status, http_status, headers, body, output):
        super().__init__(status, output)
        self.http_status = http_status
        self.headers = headers
        self.body = body

    @property
    def data(self):
        return {
            'http_status': self.http_status,
            'headers': self.headers,
            'body': self.body,
        }

    @classmethod
    def done(cls, http_status=200, headers=None, body=None):
        """
        Returns a response as the extension has successfully
        processed this interactive event.

        **Parameters:**

        * **http_status** - Optional http status to return to the
            caller (default 200 -> ok).
        * **headers** - Optional response headers to return to the
            caller.
        ** **body** - Optional response body to return to the caller.
        """
        return cls(ResultType.SUCCESS, http_status, headers, body, None)

    @classmethod
    def fail(cls, http_status=400, headers=None, body=None, output=None):
        """
        Returns a response as the extension has failed to
        process this interactive event.

        **Parameters:**

        * **http_status** - Optional http status to return to the
            caller (default 400 -> bad request).
        * **headers** - Optional response headers to return to the
            caller.
        ** **body** - Optional response body to return to the caller.
        ** **output** - Optional output message to set within the task.
        """
        return cls(ResultType.FAIL, http_status, headers, body, output)


class ValidationResponse(InteractiveResponse):
    def __init__(self, status, data, output=None):
        http_status = 200 if status == ResultType.SUCCESS else 400
        super().__init__(status, http_status, None, data, output)

    @classmethod
    def done(cls, data):
        return cls(ResultType.SUCCESS, data)

    @classmethod
    def fail(cls, data=None, output=None):
        return cls(ResultType.FAIL, data=data, output=output)


class CustomEventResponse(InteractiveResponse):
    pass


class ProductActionResponse(InteractiveResponse):
    pass


class ScheduledExecutionResponse(_Response):
    pass



import inspect
import re

from connect.eaas.core.constants import ANVIL_CALLABLE_ATTR_NAME
from connect.eaas.core.extension import AnvilApplicationBase
from connect.eaas.core.validation.models import ValidationItem, ValidationResult


def validate_anvilapp(context):

    messages = []

    if 'anvil' not in context['extension_classes']:
        return ValidationResult()

    extension_class = context['extension_classes']['anvil']
    extension_class_file = inspect.getsourcefile(extension_class)

    if not issubclass(extension_class, AnvilApplicationBase):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    f'The application class *{extension_class.__name__}* '
                    f'is not a subclass of *connect.eaas.core.extension.AnvilApplicationBase*.'
                ),
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    anvil_key_var = extension_class.get_anvil_key_variable()

    if anvil_key_var:
        variable_name_pattern = r'^[A-Za-z](?:[A-Za-z0-9_\-.]+)*$'
        variable_name_regex = re.compile(variable_name_pattern)

        if not variable_name_regex.match(anvil_key_var):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'Invalid Anvil key variable name: the value *{anvil_key_var}* '
                        f'does not match the pattern *{variable_name_pattern}*.'
                    ),
                    file=extension_class_file,
                ),
            )

    has_callables = False
    for _, value in inspect.getmembers(extension_class):
        if hasattr(value, ANVIL_CALLABLE_ATTR_NAME):
            has_callables = True
            break

    if not has_callables:
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    'The Anvil application class must contain at least one callable '
                    'marked with the *@anvil_callable* decorator.'
                ),
                file=extension_class_file,
            ),
        )
    return ValidationResult(items=messages, must_exit=len(messages) > 0)



import importlib
import inspect
import os
import re
import sys

import toml
import yaml

from connect.eaas.core.validation.helpers import get_code_context
from connect.eaas.core.validation.models import ValidationItem, ValidationResult


def _load_project_toml_file(path):
    descriptor_file = os.path.join(path, 'pyproject.toml')
    if not os.path.isfile(descriptor_file):
        return None, [
            ValidationItem(
                level='ERROR',
                message=(
                    'The mandatory *pyproject.toml* project '
                    f'descriptor file is not present in the folder {path}.'
                ),
                file=descriptor_file,
            ),
        ]
    try:
        return toml.load(descriptor_file), []
    except toml.TomlDecodeError:
        return None, [
            ValidationItem(
                level='ERROR',
                message=(
                    'The project descriptor file *pyproject.toml* is '
                    'not a valid toml file.'
                ),
                file=descriptor_file,
            ),
        ]


def validate_pyproject_toml(context):  # noqa: CCR001
    project_dir = context['project_dir']

    data, messages = _load_project_toml_file(project_dir)
    if not data:
        return ValidationResult(items=messages, must_exit=True)

    messages = []

    descriptor_file = os.path.join(project_dir, 'pyproject.toml')
    dependencies = data['tool']['poetry']['dependencies']

    if 'connect-extension-runner' in dependencies:
        messages.append(
            ValidationItem(
                message=(
                    'Extensions must depend on *connect-eaas-core* library not '
                    '*connect-extension-runner*.'
                ),
                file=descriptor_file,
            ),
        )
    elif 'connect-eaas-core' not in dependencies:
        messages.append(
            ValidationItem(
                level='ERROR',
                message='No dependency on *connect-eaas-core* has been found.',
                file=descriptor_file,
            ),
        )

    extension_dict = data['tool']['poetry'].get('plugins', {}).get('connect.eaas.ext')
    if not isinstance(extension_dict, dict):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    'No extension declaration has been found. '
                    'The extension must be declared in the '
                    '*[tool.poetry.plugins."connect.eaas.ext"]* section.'
                ),
                file=descriptor_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    sys.path.append(os.path.join(os.getcwd(), project_dir))
    possible_extensions = ['extension', 'webapp', 'anvilapp', 'eventsapp']
    extensions = {}
    for extension_type in possible_extensions:
        if extension_type in extension_dict.keys():
            package, class_name = extension_dict[extension_type].rsplit(':', 1)
            try:
                extension_module = importlib.import_module(package)
            except ImportError as err:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'The application class *{extension_dict[extension_type]}* '
                            f'cannot be loaded: {err}.'
                        ),
                        file=descriptor_file,
                    ),
                )
                return ValidationResult(items=messages, must_exit=True)

            extensions[extension_type] = getattr(extension_module, class_name)

    if 'extension' in extension_dict:
        messages.append(
            ValidationItem(
                level='WARNING',
                message=(
                    'Declaring an events application using the *extension* entrypoint name is '
                    'deprecated in favor of *eventsapp*.'
                ),
                file=descriptor_file,
            ),
        )

    if not extensions:
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    'Invalid application declaration in '
                    '*[tool.poetry.plugins."connect.eaas.ext"]*: '
                    'The application must be declared as: *"eventsapp" = '
                    '"your_package.extension:YourApplication"* '
                    'for Fulfillment automation or Hub integration. '
                    'For Multi account installation must be '
                    'declared at least one the following: *"eventsapp" = '
                    '"your_package.events:YourEventsApplication"*, '
                    '*"webapp" = "your_package.webapp:YourWebApplication"*, '
                    '*"anvilapp" = "your_package.anvil:YourAnvilApplication"*.'
                ),
                file=descriptor_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    context['extension_classes'] = extensions
    return ValidationResult(items=messages, context=context)


def validate_docker_compose_yml(context):  # noqa: CCR001
    project_dir = context['project_dir']
    messages = []
    compose_file = os.path.join(project_dir, 'docker-compose.yml')
    if not os.path.isfile(compose_file):
        messages.append(
            ValidationItem(
                message=(
                    f'The directory *{project_dir}* does not look like an extension project '
                    'directory, the file *docker-compose.yml* is not present.'
                ),
                file=compose_file,
            ),
        )
        return ValidationResult(items=messages)
    try:
        data = yaml.safe_load(open(compose_file, 'r'))
    except yaml.YAMLError:
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The file *docker-compose.yml* is not valid.',
                file=compose_file,
            ),
        )
        return ValidationResult(items=messages)

    runner_image = f'cloudblueconnect/connect-extension-runner:{context["runner_version"]}'

    for service in data['services']:
        image = data['services'][service].get('image')
        if image:
            if (
                image.startswith('cloudblueconnect/connect-extension-runner')
                and image != runner_image
            ):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid image for service *{service}*: '
                            f'expected *{runner_image}* got *{image}*.'
                        ),
                        file=compose_file,
                    ),
                )
        else:
            dockerfile = os.path.join(
                project_dir,
                data['services'][service]['build'].get('dockerfile', 'Dockerfile'),
            )
            if not os.path.isfile(dockerfile):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'The service *{service}* of *docker-compose.yml* points to a '
                            'Dockerfile that does not exist.'
                        ),
                        file=dockerfile,
                    ),
                )
                continue
            content = open(dockerfile, 'r').read().splitlines()
            from_cmd = next(filter(lambda x: x.startswith('FROM'), content), None)
            if not from_cmd:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid *Dockerfile* for service *{service}*, no '
                            'FROM statement has been found.'
                        ),
                        file=dockerfile,
                    ),
                )
                continue
            image = from_cmd[4:].strip()
            if image != runner_image:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid base image in Dockerfile of service *{service}*: '
                            f'expected *{runner_image}* got *{image}*.'
                        ),
                        file=dockerfile,
                    ),
                )

    return ValidationResult(items=messages)


def validate_variables(context):  # noqa: CCR001

    messages = []

    for _, extension_class in context['extension_classes'].items():

        variables = extension_class.get_variables()
        variable_name_pattern = r'^[A-Za-z](?:[A-Za-z0-9_\-.]+)*$'
        variable_name_regex = re.compile(variable_name_pattern)

        names = []

        for variable in variables:
            if 'name' not in variable:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message='Invalid variable declaration: the *name* attribute is mandatory.',
                        **get_code_context(extension_class, '@variables'),
                    ),
                )
                continue

            if variable["name"] in names:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Duplicate variable name: the variable with name *{variable["name"]}* '
                            'has already been declared.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )

            names.append(variable["name"])

            if not variable_name_regex.match(variable['name']):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid variable name: the value *{variable["name"]}* '
                            f'does not match the pattern *{variable_name_pattern}*.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )
            if 'initial_value' in variable and not isinstance(variable['initial_value'], str):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid *initial_value* attribute for variable *{variable["name"]}*: '
                            f'must be a non-null string not *{type(variable["initial_value"])}*.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )

            if 'secure' in variable and not isinstance(variable['secure'], bool):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid *secure* attribute for variable *{variable["name"]}*: '
                            f'must be a boolean not *{type(variable["secure"])}*.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )

    return ValidationResult(items=messages)


def validate_extension_json(context):
    messages = []
    extension_class = list(context['extension_classes'].values())[0]
    extension_class_file = inspect.getsourcefile(extension_class)
    extension_json_file = os.path.join(
        os.path.dirname(extension_class_file),
        'extension.json',
    )
    descriptor = None
    try:
        descriptor = extension_class.get_descriptor()
    except FileNotFoundError:
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The extension descriptor *extension.json* cannot be loaded.',
                file=extension_json_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    roles = ['distributor', 'reseller', 'vendor']
    if (
            not descriptor.get('audience')
            or any(role not in roles for role in descriptor['audience'])
    ):
        messages.append(
            ValidationItem(
                level='ERROR' if 'webapp' in context['extension_classes'] else 'WARNING',
                message=(
                    'For Multi account extensions the attribute *audience* that list '
                    'the target account types for this extension is mandatory in the '
                    '*extension.json* descriptor file. '
                    'Valid values are: *vendor*, *distributor*, *reseller*.'
                ),
                file=extension_json_file,
            ),
        )

    return ValidationResult(items=messages, context={
        'descriptor': descriptor,
        'extension_json_file': extension_json_file,
    })



import pytest

from connect.eaas.core.testing import WebAppTestClient


@pytest.fixture
def test_client_factory():
    """
    This fixture allows to instantiate a WebAppTestClient
    given a webapp class.
    """

    def _get_client(webapp, base_url='https://example.org/public/v1'):
        return WebAppTestClient(webapp, base_url=base_url)

    return _get_client



import inspect

from connect.eaas.core.extension import (
    EventsApplicationBase,
    Extension,
)
from connect.eaas.core.responses import (
    CustomEventResponse,
    ProcessingResponse,
    ProductActionResponse,
    ValidationResponse,
)
from connect.eaas.core.validation.helpers import get_code_context
from connect.eaas.core.validation.models import ValidationItem, ValidationResult


def validate_eventsapp(context):

    extension_class = context['extension_classes'].get(
        'eventsapp',
        context['extension_classes'].get('extension'),
    )

    if not extension_class:
        return ValidationResult()

    extension_class_file = inspect.getsourcefile(extension_class)

    messages = []

    if issubclass(extension_class, Extension):
        messages.append(
            ValidationItem(
                level='WARNING',
                message=(
                    f'The application class *{extension_class.__name__}* inherits from '
                    '*connect.eaas.extension.Extension* that has been deprecated in favor of '
                    '*connect.eaas.core.extension.EventsApplicationBase*.'
                ),
                file=extension_class_file,
            ),
        )
    elif not issubclass(extension_class, (Extension, EventsApplicationBase)):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    f'The application class *{extension_class.__name__}* '
                    'is not a subclass of *connect.eaas.core.extension.EventsApplicationBase*.'
                ),
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    defined_classes = [
        member[1]
        for member in inspect.getmembers(extension_class, predicate=inspect.isclass)
    ]

    for deprecated_cls, cls_name in (
        (CustomEventResponse, 'connect.eaas.core.responses.InteractiveResponse'),
        (ProcessingResponse, 'connect.eaas.core.responses.BackgroundResponse'),
        (ProductActionResponse, 'connect.eaas.core.responses.InteractiveResponse'),
        (ValidationResponse, 'connect.eaas.core.responses.InteractiveResponse'),
    ):
        if deprecated_cls in defined_classes:
            messages.append(
                ValidationItem(
                    message=(
                        f'The response class *{deprecated_cls.__name__}* '
                        f'has been deprecated in favor of *{cls_name}*.'
                    ),
                    **get_code_context(extension_class, deprecated_cls.__name__),
                ),
            )

    descriptor = context['descriptor']
    extension_json_file = context['extension_json_file']

    for description in ['variables', 'capabilities', 'schedulables']:
        if description in descriptor:
            messages.append(
                ValidationItem(
                    message=(
                        f'Extension {description} must be declared using the '
                        f'*connect.eaas.core.decorators.'
                        f'{description if description != "schedulables" else "event"}* decorator.'
                    ),
                    file=extension_json_file,
                ),
            )

    messages.extend(_validate_events(context))
    messages.extend(_validate_schedulables(context))

    return ValidationResult(items=messages)


def _validate_events(context):
    messages = []

    extension_class = context['extension_classes'].get(
        'eventsapp',
        context['extension_classes'].get('extension'),
    )

    definitions = context['event_definitions']
    events = extension_class.get_events()
    for event in events:
        method = getattr(extension_class, event['method'])
        if event['event_type'] not in definitions:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=f'The event type *{event["event_type"]}* is not valid.',
                    **get_code_context(method, '@event'),
                ),
            )
            continue

        statuses = definitions[event['event_type']]['object_statuses']
        if statuses:
            invalid_statuses = set(event['statuses']) - set(statuses)
        else:
            invalid_statuses = set(event['statuses'] or [])
        if invalid_statuses:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The status/es *{", ".join(invalid_statuses)}* are invalid '
                        f'for the event *{event["event_type"]}*.'
                    ),
                    **get_code_context(method, '@event'),
                ),
            )

        signature = inspect.signature(method)
        if len(signature.parameters) != 2:
            sig_str = f'{event["method"]}({", ".join(signature.parameters)})'

            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The handler for the event *{event["event_type"]}* '
                        f'has an invalid signature: *{sig_str}*'
                    ),
                    **get_code_context(method, sig_str),
                ),
            )
    return messages


def _validate_schedulables(context):
    messages = []

    extension_class = context['extension_classes'].get(
        'eventsapp',
        context['extension_classes'].get('extension'),
    )

    schedulables = extension_class.get_schedulables()
    for schedulable in schedulables:
        method = getattr(extension_class, schedulable['method'])
        signature = inspect.signature(method)
        if len(signature.parameters) != 2:
            sig_str = f'{schedulable["method"]}({", ".join(signature.parameters)})'

            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The schedulable method *{schedulable["method"]}* '
                        f'has an invalid signature: *{sig_str}*'
                    ),
                    **get_code_context(method, sig_str),
                ),
            )
    return messages



import inspect
import os

from fastapi_utils.cbv import CBV_CLASS_KEY

from connect.eaas.core.extension import WebApplicationBase
from connect.eaas.core.validation.helpers import get_code_context
from connect.eaas.core.validation.models import ValidationItem, ValidationResult


def validate_webapp(context):  # noqa: CCR001

    messages = []

    if 'webapp' not in context['extension_classes']:
        return ValidationResult()

    extension_class = context['extension_classes']['webapp']

    extension_class_file = inspect.getsourcefile(extension_class)

    if not issubclass(extension_class, WebApplicationBase):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    f'The application class *{extension_class.__name__}* '
                    f'is not a subclass of *connect.eaas.core.extension.WebApplicationBase*.'
                ),
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    if not getattr(extension_class, CBV_CLASS_KEY, False):
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The Web application class must be wrapped in *@web_app(router)*.',
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    messages.extend(_validate_webapp_routes(context))
    messages.extend(_validate_webapp_ui_modules(context))
    return ValidationResult(items=messages, must_exit=len(messages) > 0)


def _validate_webapp_routes(context):
    extension_class = context['extension_classes']['webapp']
    extension_class_file = inspect.getsourcefile(extension_class)

    auth, no_auth = extension_class.get_routers()

    no_of_routes = len(auth.routes) + len(no_auth.routes)

    if no_of_routes == 0:
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    'The Web application class must contain at least one route '
                    'implementation function wrapped in *@router.your_method("/your_path")*.'
                ),
                file=extension_class_file,
            ),
        ]
    return []


def _check_ui_component_label(extension_class, page_name, value, code_pattern):
    if not isinstance(value, str) or not value:
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    f'The label of the {page_name} must be a non-blank string.'
                ),
                **get_code_context(extension_class, code_pattern),
            ),
        ]
    return []


def _check_ui_component_url(extension_class, page_name, value, code_pattern):
    if not isinstance(value, str) or not value:
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    f'The url of the {page_name} must be a '
                    'path to a html file relative to the static folder.'
                ),
                **get_code_context(extension_class, code_pattern),
            ),
        ]

    if not value.startswith('/static/'):
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    f'The url {value} of the {page_name} must be prefixed with /static.'
                ),
                **get_code_context(extension_class, code_pattern),
            ),
        ]

    page_path = value[8:]
    full_path = os.path.join(
        os.path.dirname(inspect.getsourcefile(extension_class)),
        'static',
        page_path,
    )
    if not os.path.exists(full_path):
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    f'The url {value} of the {page_name} page does not point to any file.'
                ),
                **get_code_context(extension_class, code_pattern),
            ),
        ]
    return []


def _validate_webapp_ui_modules(context):  # noqa: CCR001
    messages = []
    extension_class = context['extension_classes']['webapp']

    ui_modules = extension_class.get_ui_modules()

    if 'settings' in ui_modules:
        label = ui_modules['settings']['label']
        url = ui_modules['settings']['url']

        messages.extend(
            _check_ui_component_label(
                extension_class, '"Account Settings"', label, '@account_settings_page(',
            ),
        )

        messages.extend(
            _check_ui_component_url(
                extension_class, '"Account Settings"', url, '@account_settings_page(',
            ),
        )

    if 'modules' in ui_modules:
        label = ui_modules['modules']['label']
        url = ui_modules['modules']['url']
        messages.extend(
            _check_ui_component_label(
                extension_class, '"Module Main Page"', label, '@module_pages(',
            ),
        )
        messages.extend(
            _check_ui_component_url(
                extension_class, '"Module Main Page"', url, '@module_pages(',
            ),
        )

        if 'children' in ui_modules['modules']:
            children = ui_modules['modules']['children']
            if not isinstance(children, (list, tuple)):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            'The "children" argument of the @module_pages '
                            'decorator must be a list of objects.'
                        ),
                        **get_code_context(extension_class, '@module_pages('),
                    ),
                )
            else:
                for child in children:
                    if 'label' not in child or 'url' not in child:
                        messages.append(
                            ValidationItem(
                                level='ERROR',
                                message=(
                                    'Invalid child page declaration. '
                                    'Each child page must be an object with the '
                                    'label and url attributes.'
                                ),
                                **get_code_context(extension_class, '@module_pages('),
                            ),
                        )
                        continue

                    label = child['label']
                    url = child['url']
                    messages.extend(
                        _check_ui_component_label(
                            extension_class, '"Module Child Page"', label, '@module_pages(',
                        ),
                    )
                    messages.extend(
                        _check_ui_component_url(
                            extension_class, '"Module Child Page"', url, '@module_pages(',
                        ),
                    )

    if 'admins' in ui_modules:
        admins = ui_modules['admins']
        if not isinstance(admins, (list, tuple)):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The argument of the @admin_pages '
                        'decorator must be a list of objects.'
                    ),
                    **get_code_context(extension_class, '@admin_pages('),
                ),
            )
        else:
            for child in admins:
                if 'label' not in child or 'url' not in child:
                    messages.append(
                        ValidationItem(
                            level='ERROR',
                            message=(
                                'Invalid admin page declaration. '
                                'Each admin page must be an object with the '
                                'label and url attributes.'
                            ),
                            **get_code_context(extension_class, '@admin_pages('),
                        ),
                    )
                    continue

                label = child['label']
                url = child['url']
                messages.extend(
                    _check_ui_component_label(
                        extension_class, '"Admin Page"', label, '@admin_pages(',
                    ),
                )
                messages.extend(
                    _check_ui_component_url(
                        extension_class, '"Admin Page"', url, '@admin_pages(',
                    ),
                )

    return messages



import functools
import inspect
import json
import os

import anvil.server
import pkg_resources
from fastapi import APIRouter

from connect.client import AsyncConnectClient, ConnectClient
from connect.eaas.core.constants import (
    ACCOUNT_SETTINGS_PAGE_ATTR_NAME,
    ADMIN_PAGES_ATTR_NAME,
    ANVIL_CALLABLE_ATTR_NAME,
    ANVIL_KEY_VAR_ATTR_NAME,
    EVENT_INFO_ATTR_NAME,
    GUEST_ENDPOINT_ATTR_NAME,
    MODULE_PAGES_ATTR_NAME,
    SCHEDULABLE_INFO_ATTR_NAME,
    TRANSFORMATION_ATTR_NAME,
    VARIABLES_INFO_ATTR_NAME,
)
from connect.eaas.core.decorators import router


class ApplicationBase:

    @classmethod
    def get_descriptor(cls):  # pragma: no cover
        """
        Returns the **extension.json** extension descriptor.

        !!! example

            ``` python
            {
                'name': 'My extension',
                'description': 'This extension My Extension',
                'version': '1.0.0',
                'audience': ['distributor', 'vendor', 'reseller'],
                'readme_url': 'https://example.org/README.md',
                'changelog_url': 'https://example.org/CHANGELOG.md',
                'icon': 'bar_chart',
            }
            ```
        """
        return json.load(
            pkg_resources.resource_stream(
                cls.__module__,
                'extension.json',
            ),
        )

    @classmethod
    def get_variables(cls):
        """
        Inspects the Events Application class returning the list of environment
        variables declared using the `@variables` class decorator.

        !!! example

            ``` python
            [
                {
                    'name': 'MY_ENV_VAR',
                    'initial_value': '<<change me>>',
                    'secure': False,
                },
                {
                    'name': 'MY_SECURE_ENV_VAR',
                    'initial_value': '<<change me>>',
                    'secure': True,
                },
            ]
            ```
        """
        return getattr(cls, VARIABLES_INFO_ATTR_NAME, [])


class EventsApplicationBase(ApplicationBase):
    """Base class to implements an Events Application."""

    def __init__(
        self,
        client,
        logger,
        config,
        installation_client=None,
        installation=None,
        context=None,
    ):
        self.client = client
        self.logger = logger
        self.config = config
        self.installation_client = installation_client
        self.installation = installation
        self.context = context

    def get_installation_admin_client(self, installation_id):
        data = (
            self.client('devops')
            .services[self.context.extension_id]
            .installations[installation_id]
            .action('impersonate')
            .post()
        )

        return ConnectClient(
            data['installation_api_key'],
            endpoint=self.client.endpoint,
            default_headers=self.client.default_headers,
            logger=self.client.logger,
        )

    async def get_installation_admin_async_client(self, installation_id):
        data = (
            await self.client('devops')
            .services[self.context.extension_id]
            .installations[installation_id]
            .action('impersonate')
            .post()
        )
        return AsyncConnectClient(
            data['installation_api_key'],
            endpoint=self.client.endpoint,
            default_headers=self.client.default_headers,
            logger=self.client.logger,
        )

    @classmethod
    def get_events(cls):
        """
        Inspects the Events Application class for methods
        decorated with the `@event` decorator and return a list of
        objects like in the following example:

        !!! example

            ``` python
            [
                {
                    'method': 'handle_purchase_request',
                    'event_type': 'asset_purchase_request_processing',
                    'statuses': ['pending', 'approved'],
                },
            ]
            ```
        """
        return cls._get_methods_info(EVENT_INFO_ATTR_NAME)

    @classmethod
    def get_schedulables(cls):
        """
        Inspects the Events Application class for methods
        decorated with the `@schedulable` decorator and return a list of
        objects like in the following example:

        !!! example

            ``` python
            [
                {
                    'method': 'refresh_oauth_token',
                    'name': 'Refresh OAuth Token',
                    'description': 'This schedulable refresh the GCP OAuth toke',
                },
            ]
            ```
        """
        return cls._get_methods_info(SCHEDULABLE_INFO_ATTR_NAME)

    @classmethod
    def _get_methods_info(cls, attr_name):
        info = []
        members = inspect.getmembers(cls)
        for _, value in members:
            if not (inspect.isfunction(value) or inspect.iscoroutinefunction(value)):
                continue
            if hasattr(value, attr_name):
                info.append(getattr(value, attr_name))
        return info


class Extension(EventsApplicationBase):
    pass


class WebApplicationBase(ApplicationBase):
    """Base class to implements an Events Application."""
    @classmethod
    def get_static_root(cls):
        """Returns the absolute path to the `static` root folder."""
        static_root = os.path.abspath(
            os.path.join(
                os.path.dirname(inspect.getfile(cls)),
                'static',
            ),
        )
        if os.path.exists(static_root) and os.path.isdir(static_root):
            return static_root
        return None

    @classmethod
    def get_routers(cls):
        """
        Inspect the Web Application class for routes and return a tuple
        or two routers, the first one contains authenticated API routes,
        the second the non-authenticated ones.

        !!! warning
            Non authenticated endpoints must be authorized by CloudBlue.
            If your extension need to expose some, please contact the
            CloudBlue support.
        """
        auth = APIRouter()
        no_auth = APIRouter()
        for route in router.routes:
            if getattr(route.endpoint, GUEST_ENDPOINT_ATTR_NAME, False):
                no_auth.routes.append(route)
            else:
                auth.routes.append(route)
        return auth, no_auth

    @classmethod
    def get_ui_modules(cls):
        ui_modules = {}
        if hasattr(cls, ACCOUNT_SETTINGS_PAGE_ATTR_NAME):
            ui_modules['settings'] = getattr(cls, ACCOUNT_SETTINGS_PAGE_ATTR_NAME)

        if hasattr(cls, MODULE_PAGES_ATTR_NAME):
            ui_modules['modules'] = getattr(cls, MODULE_PAGES_ATTR_NAME)

        if hasattr(cls, ADMIN_PAGES_ATTR_NAME):
            ui_modules['admins'] = getattr(cls, ADMIN_PAGES_ATTR_NAME)

        return ui_modules


def _invoke(method, **kwargs):
    return method(**kwargs)


class AnvilApplicationBase(ApplicationBase):
    """Base class to implements an Anvil Application."""
    def __init__(self, client, logger, config, installation_client=None, installation=None):
        self.client = client
        self.logger = logger
        self.config = config
        self.installation_client = installation_client
        self.installation = installation

    @classmethod
    def get_anvil_key_variable(cls):
        """
        Returns the name of the environment variable that
        stores the Anvil Server Uplink key.
        """
        return getattr(cls, ANVIL_KEY_VAR_ATTR_NAME, [])

    @classmethod
    def get_anvil_callables(cls):
        """
        Inspect the Anvil Application class to searching for
        method decorated with the `@anvil_callable` decorator and
        return a list of discovered anvil callable like in the following
        example:

        !!! example

            ``` python
            [
                {
                    'method': 'my_anvil_callable',
                    'summary': 'This callable sum two integer.',
                    'description': 'Example anvil callable.',
                },
            ]
            ```
        """
        callables = []
        members = inspect.getmembers(cls)
        for _, value in members:
            if not inspect.isfunction(value):
                continue

            if hasattr(value, ANVIL_CALLABLE_ATTR_NAME):
                callables.append(getattr(value, ANVIL_CALLABLE_ATTR_NAME))
        return callables

    def setup_anvil_callables(self):
        """
        Setup the Anvil Server Uplink instance with the discovered
        Anvil callables.
        """
        members = inspect.getmembers(self)
        for _, value in members:
            if not inspect.ismethod(value):
                continue

            if getattr(value, ANVIL_CALLABLE_ATTR_NAME, False):
                fn = functools.partial(_invoke, value)
                fn.__name__ = value.__name__
                anvil.server.callable(fn)


class TransformationBase:

    def __init__(
            self,
            input_columns,
            output_columns,
            stream,
            client,
            config,
            logger,
            transformation_settings=None,
            installation_client=None,
            installation=None,
            cache=None,
    ):
        self.client = client
        self.config = config
        self.logger = logger

        self.installation_client = installation_client
        self.installation = installation

        self.input_columns = input_columns
        self.output_columns = output_columns
        self.stream = stream

        self.transformation_settings = transformation_settings
        self.cache = cache

    @classmethod
    def get_transformation_info(cls):
        return getattr(cls, TRANSFORMATION_ATTR_NAME, None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['connect',
 'connect.eaas.core',
 'connect.eaas.core.inject',
 'connect.eaas.core.testing',
 'connect.eaas.core.validation',
 'connect.eaas.core.validation.validators']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'anvil-uplink>=0.4.0,<0.5.0',
 'connect-openapi-client>=25.15',
 'fastapi-utils>=0.2.1,<0.3.0',
 'fastapi>=0.78.0,<0.79.0',
 'logzio-python-handler>=3.1.1,<4.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'pytest11': ['pytest_eaas_core = connect.eaas.core.testing.fixtures']}

setup_kwargs = {
    'name': 'connect-eaas-core',
    'version': '27.1',
    'description': 'Connect Eaas Core',
    'long_description': '# Connect Extension as a Service Core Library\n\n![pyversions](https://img.shields.io/pypi/pyversions/connect-eaas-core.svg) [![PyPi Status](https://img.shields.io/pypi/v/connect-eaas-core.svg)](https://pypi.org/project/connect-eaas-core/) [![Build Status](https://github.com/cloudblue/connect-eaas-core/workflows/Build%20Connect%20EaaS%20Core/badge.svg)](https://github.com/cloudblue/connect-eaas-core/actions) [![Sonarcloud Status](https://sonarcloud.io/api/project_badges/measure?project=connect-eaas-core&metric=alert_status)](https://sonarcloud.io/dashboard?id=connect-eaas-core) \n [![SonarCloud Coverage](https://sonarcloud.io/api/project_badges/measure?project=connect-eaas-core&metric=coverage)](https://sonarcloud.io/component_measures/metric/coverage/list?id=connect-eaas-core)\n [![SonarCloud Bugs](https://sonarcloud.io/api/project_badges/measure?project=connect-eaas-core&metric=bugs)](https://sonarcloud.io/component_measures/metric/reliability_rating/list?id=connect-eaas-core)\n [![SonarCloud Vulnerabilities](https://sonarcloud.io/api/project_badges/measure?project=connect-eaas-core&metric=vulnerabilities)](https://sonarcloud.io/component_measures/metric/security_rating/list?id=connect-eaas-core)\n\n\n\n\n## Introduction\n\n`Connect EaaS Core` is a library that contains the base building blocks to develop extensions for the CloudBlue Connect platform.\n\n\n## Install\n\n`Connect EaaS Core` requires python 3.8 or later.\n\n\n`Connect EaaS Core` can be installed from [pypi.org](https://pypi.org/project/connect-eaas-core/) using pip:\n\n```\n$ pip install connect-eaas-core\n```\n\n\n## Documentation\n\n[`Connect EaaS Core` documentation](https://connect-eaas-core.readthedocs.io/en/latest/) is hosted on the _Read the Docs_ service.\n\n\n## License\n\n`Connect EaaS Core` is released under the [Apache License Version 2.0](https://www.apache.org/licenses/LICENSE-2.0).\n',
    'author': 'CloudBlue LLC',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from fastapi_utils.cbv import cbv
from fastapi_utils.inferring_router import InferringRouter

from connect.eaas.core.constants import (
    ACCOUNT_SETTINGS_PAGE_ATTR_NAME,
    ADMIN_PAGES_ATTR_NAME,
    ANVIL_CALLABLE_ATTR_NAME,
    ANVIL_KEY_VAR_ATTR_NAME,
    EVENT_INFO_ATTR_NAME,
    GUEST_ENDPOINT_ATTR_NAME,
    MODULE_PAGES_ATTR_NAME,
    SCHEDULABLE_INFO_ATTR_NAME,
    TRANSFORMATION_ATTR_NAME,
    VARIABLES_INFO_ATTR_NAME,
)


def event(event_type, statuses=None):
    """
    Mark a method of an Events Application as the handler
    for a given `event_type` eventually filtering the event
    by status.

    Usage:

    ``` python
    from connect.eaas.core.decorators import event
    from connect.eaas.core.extension import EventsApplicationBase


    class MyEventsApplication(EventsApplicationBase):

        @event(
            'asset_purchase_request_processing',
            statuses=[
                'pending', 'approved', 'failed',
                'inquiring', 'scheduled', 'revoking',
                'revoked',
            ],
        )
        def handle_purchase_request(self, request):
            pass
    ```

    **Parameters:**

    * **event_type** - The type of event this handler is for.
    * **statuses** - List of statuses of the event that this handler
    want to receive.

    !!! note
        The list of statuses is required for `background` event types
        only and must not be set for `interactive` events.
    """
    def wrapper(func):
        setattr(
            func,
            EVENT_INFO_ATTR_NAME,
            {
                'method': func.__name__,
                'event_type': event_type,
                'statuses': statuses,
            },
        )
        return func
    return wrapper


def schedulable(name, description):
    """
    Mark a method of an Events Application as that can be invoked
    on a scheduled basis.

    Usage:

    ``` python
    from connect.eaas.core.decorators import schedulable
    from connect.eaas.core.extension import EventsApplicationBase

    class MyEventsApplication(EventsApplicationBase):

        @schedulable(
            'Schedulable method mock',
            'It can be used to test DevOps scheduler.',
        )
        def execute_scheduled_processing(self, schedule):
            pass
    ```

    **Parameters:**

    * **name** - The name of this schedulable method.
    * **description** - Description of what this schedulable method do.

    !!! note
        The `name` and `description` arguments are used by the
        `create new schedule` wizard of the Connect DevOps module
        to allow to identify the method when creating schedules for it.
    """
    def wrapper(func):
        setattr(
            func,
            SCHEDULABLE_INFO_ATTR_NAME,
            {
                'method': func.__name__,
                'name': name,
                'description': description,
            },
        )
        return func
    return wrapper


def variables(variables):
    """
    Class decorator to declare variables needed by your application.
    The declared variables will be created on the first run with the
    specified `initial_value` if they don't exist.

    Usage:

    ``` python
    from connect.eaas.core.decorators import variables
    from connect.eaas.core.extension import EventsApplicationBase


    @variables(
        [
            {
            "name": "ASSET_REQUEST_APPROVE_TEMPLATE_ID",
            "initial_value": "<change_with_purchase_request_approve_template_id>"
            },
            {
            "name": "ASSET_REQUEST_CHANGE_TEMPLATE_ID",
            "initial_value": "<change_with_change_request_approve_template_id>"
            },
            {
            "name": "TIER_REQUEST_APPROVE_TEMPLATE_ID",
            "initial_value": "<change_with_tier_request_approve_template_id>"
            },
        ],
    )
    class MyEventsApplication(EventsApplicationBase):
        pass
    ```

    **Parameters:**

    * **variables** - The list of environment variables you want to initialize.
    """
    def wrapper(cls):
        if hasattr(cls, VARIABLES_INFO_ATTR_NAME):
            declared_vars = getattr(cls, VARIABLES_INFO_ATTR_NAME)
            for variable in variables:
                name = variable['name']
                if len(list(filter(lambda x: x['name'] == name, declared_vars))) == 0:  # noqa: B023
                    declared_vars.append(variable)
            return cls
        setattr(cls, VARIABLES_INFO_ATTR_NAME, variables)
        return cls
    return wrapper


def anvil_key_variable(name):
    """
    Class decorator for an Anvil Application that declare the name
    of the environment variable that stores the Anvil Server Uplink
    key.

    Usage:

    ``` python
    from connect.eaas.core.decorators import anvil_key_variable
    from connect.eaas.core.extension import AnvilApplicationBase


    @anvil_key_variable('ANVIL_SERVER_UPLINK_KEY')
    class MyAnvilApplication(AnvilApplicationBase):
        pass
    ```

    **Parameters:**

    * **name** - Name of the environment variable.

    !!! note
        This environment variable does not need to be
        declared using the `@variables` decorator it will be
        automatically declared as a `secure` variable with
        the `initial_value` set to `changeme!`.
    """
    def wrapper(cls):
        setattr(cls, ANVIL_KEY_VAR_ATTR_NAME, name)
        variables = []
        if not hasattr(cls, VARIABLES_INFO_ATTR_NAME):
            setattr(cls, VARIABLES_INFO_ATTR_NAME, variables)
        else:
            variables = getattr(cls, VARIABLES_INFO_ATTR_NAME)
        if len(list(filter(lambda x: x['name'] == name, variables))) == 0:
            variables.append({'name': name, 'initial_value': 'changeme!', 'secure': True})
        return cls
    return wrapper


def anvil_callable(summary=None, description=None):
    """
    Mark a method of an Anvil Application class as a method that
    can be called from an Anvil frontend application.

    Usage:

    ``` python
    from connect.eaas.core.decorators import anvil_callable, anvil_key_variable
    from connect.eaas.core.extension import AnvilApplicationBase


    @anvil_key_variable('ANVIL_SERVER_UPLINK_KEY')
    class MyAnvilApplication(AnvilApplicationBase):

        @anvil_callable(
            summary='This function say hello',
            decription='Description of what this function do',
        )
        def say_hello(self, name):
            return f'Hello {name}'
    ```
    """
    def wrapper(func):
        setattr(
            func,
            ANVIL_CALLABLE_ATTR_NAME,
            {
                'method': func.__name__,
                'summary': summary or func.__name__.replace('_', ' ').title(),
                'description': description or func.__doc__ or '',
            },
        )
        return func
    return wrapper


def guest():
    """
    Mark an endpoint of a Web Application as not subject to
    the Connect authentication.

    Usage:

    ``` python
    from connect.eaas.core.decorators import guest, router, web_app
    from connect.eaas.core.extension import WebApplicationBase


    @web_app(router)
    class MyWebApp(WebApplicationBase):

        @guest()
        @router.get('/my_endpoint')
        def my_endpoint(self):
            pass
    ```

    !!! warning
        Non authenticated endpoints must be authorized by CloudBlue.
        If your extension need to expose some, please contact the
        CloudBlue support.
    """
    def wrapper(func):
        setattr(func, GUEST_ENDPOINT_ATTR_NAME, True)
        return func
    return wrapper


def account_settings_page(label, url):
    """
    Class decorator for Web Application that declare which html page
    must be rendererd within the `Account Settings` module of Connect UI
    to handle the extension installation settings.

    Usage:

    ``` python
    from connect.eaas.core.decorators import (
        account_settings_page, router, web_app,
    )
    from connect.eaas.core.extension import WebApplicationBase


    @web_app(router)
    @account_settings_page('My settings', '/static/settings.html')
    class MyWebApp(WebApplicationBase):
        pass
    ```

    **Parameters:**

    * **label** - the label to use for such page.
    * **url** - the url path to the html page including `/static`.
    """
    def wrapper(cls):
        setattr(cls, ACCOUNT_SETTINGS_PAGE_ATTR_NAME, {'label': label, 'url': url})
        return cls
    return wrapper


def module_pages(label, url, children=None):
    """
    Class decorator for Web Application that declare the main page
    for a web application and optionally a list of children pages.

    Usage:

    ``` python
    from connect.eaas.core.decorators import (
        module_pages, router, web_app,
    )
    from connect.eaas.core.extension import WebApplicationBase


    @web_app(router)
    @module_pages(
        'Home',
        '/static/home.html',
        children=[
            {
                'label': 'Child page 1',
                'url': '/static/child1.html',
            },
        ],
    )
    class MyWebApp(WebApplicationBase):
        pass
    ```

    **Parameters:**

    * **label** - the label to use for such page.
    * **url** - the url path to the html page including `/static`.
    * **children** - optional list of children pages.

    !!! note
        Your extension pages will be rendered using a `Tabs` component
        of the Connect UI.
        The main page will be the first tab and the label will be the tab
        label. For each child in children a tab will be added to the `Tabs`
        component using its label as the tab label.
    """
    def wrapper(cls):
        data = {
            'label': label,
            'url': url,
        }
        if children:
            data['children'] = children
        setattr(cls, MODULE_PAGES_ATTR_NAME, data)
        return cls
    return wrapper


def admin_pages(pages):
    """
    Class decorator for Web Application that declare a list of
    admin pages.
    Admin pages are shown in the detail view of a specific installation
    of the Web Application inside your extension details view.

    Usage:

    ``` python
    from connect.eaas.core.decorators import (
        admin_pages, router, web_app,
    )
    from connect.eaas.core.extension import WebApplicationBase


    @web_app(router)
    @admin_pages(
        [
            {
                'label': 'My Admin 1',
                'url': '/static/admin1.html',
            },
        ],
    )
    class MyWebApp(WebApplicationBase):
        pass
    ```

    **Parameters:**

    * **pages** - optional list of admin pages.
    """
    def wrapper(cls):
        setattr(cls, ADMIN_PAGES_ATTR_NAME, pages)
        return cls
    return wrapper


def transformation(name, description, edit_dialog_ui):
    def wrapper(cls):
        setattr(cls, TRANSFORMATION_ATTR_NAME, {
            'name': name,
            'description': description,
            'edit_dialog_ui': edit_dialog_ui,
            'class_fqn': f'{cls.__module__}.{cls.__name__}',
        })
        return cls
    return wrapper


router = InferringRouter()
web_app = cbv



import inspect


def get_code_context(obj, pattern):
    source_lines = inspect.getsourcelines(obj)
    file = inspect.getsourcefile(obj)
    start_line = source_lines[1]
    lineno = source_lines[1]
    code = ''.join(source_lines[0])

    for idx, line in enumerate(source_lines[0]):  # pragma: no branch
        if pattern in line:
            lineno += idx
            break

    if inspect.ismodule(obj):
        code = ''.join(code.splitlines(keepends=True)[0:lineno + 3])

    return {
        'file': file,
        'start_line': start_line,
        'lineno': lineno,
        'code': code,
    }



import inspect
import json
from urllib.parse import urlparse

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.routing import Match
from starlette.testclient import TestClient

from connect.client import ClientError
from connect.client.testing import AsyncConnectClientMocker, ConnectClientMocker
from connect.eaas.core.inject.models import Context
from connect.eaas.core.utils import client_error_exception_handler


class WebAppTestClient(TestClient):
    """
    Create a new instance of the WebAppTestClient.
    The WebAppTestClient is based on the python requests library
    and will internally create a requests.Session.
    Each interface method (get, post, ...) accepts the following extra
    keyword arguments:

    * **context**: an instance of connect.eaas.core.inject.models.Context
        to provide the endpoint with a specific call context or a dict that will merged
        with a default context.
        If this kwarg is omitted a default context is provided.
    * **installation**: an installation object to receive inside the called endpoint,
        defaults to None.
    * **config**: a dictionary with key/value pairs representing environment variables.
    * **log_level**: allow to simulate a specific log level for such endpoint call
        default to INFO.


    Usage:

    ```python
    from connect.eaas.core.testing.testclient import WebAppTestClient

    from myext.webapp import MyWebApplication

    client = WebAppTestClient(MyWebApplication)
    response = client.get('/my/endpoint')
    ```

    **Parameters:**

    * **webapp** - The web application class.
    """
    def __init__(self, webapp, base_url='https://example.org/public/v1'):
        self._webapp_class = webapp
        self._app = self._get_application()

        super().__init__(app=self._app, base_url=base_url)

        self.headers = {
            'X-Connect-Api-Gateway-Url': self.base_url,
            'X-Connect-User-Agent': 'eaas-test-client',
            'X-Connect-Installation-Api-Key': 'ApiKey XXXX',
        }

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
        context=None,
        installation=None,
        config=None,
        log_level=None,
    ):
        headers = self._populate_internal_headers(
            headers or {},
            context=context,
            installation=installation,
            config=config,
            log_level=log_level,
        )
        mocker = self._get_client_mocker(method, url)
        if installation and mocker:
            with mocker(self.base_url) as mocker:
                mocker.ns('devops').collection('installations').resource(
                    installation['id'],
                ).get(return_value=installation)
                return super().request(
                    method,
                    url,
                    params=params,
                    data=data,
                    headers=headers,
                    cookies=cookies,
                    files=files,
                    auth=auth,
                    timeout=timeout,
                    allow_redirects=allow_redirects,
                    proxies=proxies,
                    hooks=hooks,
                    stream=stream,
                    verify=verify,
                    cert=cert,
                    json=json,
                )
        return super().request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )

    def _get_client_mocker(self, method, url):
        path = urlparse(url).path
        for route in self.app.router.routes:
            match, child_scope = route.matches({'type': 'http', 'method': method, 'path': path})
            if match == Match.FULL:
                if inspect.iscoroutinefunction(child_scope['endpoint']):
                    return AsyncConnectClientMocker
                return ConnectClientMocker

    def _generate_call_context(self, installation):
        return {
            'extension_id': 'SRVC-0000',
            'environment_id': 'ENV-0000-03',
            'environment_type': 'production',
            'installation_id': installation['id'] if installation else None,
            'user_id': 'UR-000',
            'account_id': 'VA-000',
            'account_role': 'vendor',
            'call_source': 'ui',
            'call_type': 'user',
        }

    def _populate_internal_headers(
        self,
        headers,
        config=None,
        installation=None,
        context=None,
        log_level=None,
    ):

        if not context:
            context = Context(**self._generate_call_context(installation))
        elif isinstance(context, dict):
            ctx_data = self._generate_call_context(installation)
            ctx_data.update(context)
            context = Context(**ctx_data)

        headers['X-Connect-Logging-Level'] = log_level or 'INFO'
        if config:
            headers['X-Connect-Config'] = json.dumps(config)
        if context.installation_id:
            headers['X-Connect-Installation-id'] = context.installation_id
        headers['X-Connect-Extension-Id'] = context.extension_id
        headers['X-Connect-Environment-Id'] = context.environment_id
        headers['X-Connect-Environment-Type'] = context.environment_type
        headers['X-Connect-User-Id'] = context.user_id
        headers['X-Connect-Account-Id'] = context.account_id
        headers['X-Connect-Account-Role'] = context.account_role
        headers['X-Connect-Call-Source'] = context.call_source
        headers['X-Connect-Call-Type'] = context.call_type
        return headers

    def _get_application(self):
        app = FastAPI(
            exception_handlers={
                ClientError: client_error_exception_handler,
            },
        )
        auth_router, no_auth_router = self._webapp_class.get_routers()
        app.include_router(auth_router, prefix='/api')
        app.include_router(no_auth_router, prefix='/guest')

        static_root = self._webapp_class.get_static_root()
        if static_root:
            app.mount('/static', StaticFiles(directory=static_root), name='static')

        return app



import os
from logging import Logger

from fastapi import Depends, Header, Path

from connect.client import AsyncConnectClient
from connect.eaas.core.inject.common import get_call_context, get_logger
from connect.eaas.core.inject.models import Context
from connect.eaas.core.logging import RequestLogger
from connect.eaas.core.utils import get_correlation_id


def get_installation_client(
    logger: Logger = Depends(get_logger),
    x_connect_installation_api_key: str = Header(),
    x_connect_api_gateway_url: str = Header(),
    x_connect_user_agent: str = Header(),
    x_connect_correlation_id: str = Header(None),
):
    default_headers = {
        'User-Agent': x_connect_user_agent,
    }
    if x_connect_correlation_id:
        default_headers['ext-traceparent'] = get_correlation_id(x_connect_correlation_id)

    return AsyncConnectClient(
        x_connect_installation_api_key,
        endpoint=x_connect_api_gateway_url,
        use_specs=False,
        default_headers=default_headers,
        logger=RequestLogger(logger),
    )


def get_extension_client(
    logger: Logger = Depends(get_logger),
    x_connect_api_gateway_url: str = Header(),
    x_connect_user_agent: str = Header(),
    x_connect_correlation_id: str = Header(None),
):
    default_headers = {
        'User-Agent': x_connect_user_agent,
    }
    if x_connect_correlation_id:
        default_headers['ext-traceparent'] = get_correlation_id(x_connect_correlation_id)

    return AsyncConnectClient(
        os.getenv('API_KEY'),
        endpoint=x_connect_api_gateway_url,
        use_specs=False,
        default_headers=default_headers,
        logger=RequestLogger(logger),
    )


async def get_installation(
    client: AsyncConnectClient = Depends(get_installation_client),
    x_connect_installation_id: str = Header(),
):
    return await client("devops").installations[x_connect_installation_id].get()


async def get_installation_admin_client(
    installation_id: str = Path(),
    context: Context = Depends(get_call_context),
    client: AsyncConnectClient = Depends(get_extension_client),
):
    data = (
        await client('devops')
        .services[context.extension_id]
        .installations[installation_id]
        .action('impersonate')
        .post()
    )

    return AsyncConnectClient(
        data['installation_api_key'],
        endpoint=client.endpoint,
        default_headers=client.default_headers,
        logger=client.logger,
    )
